package main

import (
	"context"
	"flag"
	"fmt"
	"math/bits"
	"os"
	"os/exec"
	"runtime"
	"strings"
	"time"

	"github.com/Tom-Johnston/mamba/graph"
	"github.com/Tom-Johnston/mamba/graph/search"
	"github.com/Tom-Johnston/mamba/ints"
	"github.com/Tom-Johnston/mamba/itertools"
)

func main() {

	start := time.Now()
	nPtr := flag.Int("n", -1, "enumerate the circle graphs with n vertices")
	algPtr := flag.Int("alg", 0, `specify which algorithm to use. 
	- 0: canonical deletion
	- 1: check all double occurrence words
	- 2: check all topological sorts
	- 3: check all orbits
	- 4: check all graphs`)
	comparePtr := flag.Bool("comp", false, "compare the different methods")
	flag.Parse()

	n := *nPtr
	alg := *algPtr
	if *comparePtr {
		compare()
		return
	}

	if n == -1 {
		panic("The option -n must be specified and a non-negative integer.")
	}

	runtime.GOMAXPROCS(1)

	counter := 0

	switch alg {
	case 0:
		fmt.Printf("Counting circle graphs on %v vertices using canonical deletion.\n", n)
		counter = checkCanonicalDeletion(n)
	case 1:
		fmt.Printf("Counting circle graphs on %v vertices by checking all double occurrence words.\n", n)
		counter = checkDOW(n)
	case 2:
		fmt.Printf("Counting circle graphs on %v vertices by checking all topological sorts.\n", n)
		counter = checkTopologicalSorts(n)
	case 3:
		fmt.Printf("Counting circle graphs on %v vertices by checking all orbits.\n", n)
		counter = checkOrbits(n)
	case 4:
		fmt.Printf("Counting circle graphs on %v vertices by checking all graphs.\n", n)
		counter = checkAllGraphs(n)
	}

	fmt.Println("Graphs: ", counter)
	fmt.Printf("Took %v\n", time.Since(start))
}

func compare() {
	ex, err := os.Executable()
	if err != nil {
		panic(err)
	}
	for alg := 0; alg < 5; alg++ {
		for n := 1; n < 20; n++ {
			ctx, cancel := context.WithTimeout(context.Background(), time.Hour)
			defer cancel()

			com := exec.CommandContext(ctx, ex, "-n", fmt.Sprint(n), "-alg", fmt.Sprint(alg))
			com.Stdout = os.Stdout
			err := com.Run()
			if err != nil {
				//We will assume it timed out.
				break
			}
			fmt.Println("")
		}
	}
}

func checkDOW(n int) int {
	//Iterate over all double occurrence words.
	freq := make([]int, n)
	for i := range freq {
		freq[i] = 2
	}
	iter := itertools.MultisetPermutations(freq)

	//Use a map to act as a set.
	graphs := make(map[string]struct{})

	//seen is used when computing the graph from the double occurrence word. It is allocated here and reused.
	seen := make([]bool, n)
	//g will store the circle graph correspoding to the double occurrence word. It is allocated here and reused.
	g := graph.NewDense(n, nil)
	for iter.Next() {
		//Reset the graph g.
		for i := range g.Edges {
			g.Edges[i] = 0
		}
		for i := range g.DegreeSequence {
			g.DegreeSequence[i] = 0
		}
		g.NumberOfEdges = 0

		//Compute the graph corresponding to v.
		v := iter.Value()
		for i, u := range v {
			//Reset seen
			for j := range seen {
				seen[j] = false
			}
			//Look for the next occurrence of this letter, switching seen for each letter we see.
			for j := i + 1; j < 2*n; j++ {
				w := v[j]
				if w == u {
					//We have reached the same letter again so add the appropriate edges.
					for k := range seen {
						if seen[k] {
							g.AddEdge(u, k)
						}
					}
					break
				}
				seen[w] = !seen[w]
			}
		}
		//Compute the canonical isomorph and encode it in graph6 format.
		ci := graph.CanonicalIsomorph(g)
		g6 := graph.Graph6Encode(graph.InducedSubgraph(g, ci))
		//Add the encoded version of the graph to the set.
		graphs[g6] = struct{}{}
	}
	return len(graphs)
}

func checkTopologicalSorts(n int) int {
	less := func(i, j int) bool {
		if i < j && j < n {
			return true
		}
		if i%n == j%n {
			return i < j
		}
		return false
	}

	iter := itertools.TopologicalSorts(2*n, less)

	graphs := make(map[string]struct{})
	seen := make([]bool, n)
	g := graph.NewDense(n, nil)
	for iter.Next() {
		//Reset the graph g.
		for i := range g.Edges {
			g.Edges[i] = 0
		}
		for i := range g.DegreeSequence {
			g.DegreeSequence[i] = 0
		}
		g.NumberOfEdges = 0

		v := iter.Value()
		for i, u := range v {
			for j := range seen {
				seen[j] = false
			}
			for j := i + 1; j < 2*n; j++ {
				w := v[j] % n
				if w == u {
					for k := range seen {
						if seen[k] {
							g.AddEdge(u, k)
						}
					}
					break
				}
				seen[w] = !seen[w]
			}
		}
		ci := graph.CanonicalIsomorph(g)
		g6 := graph.Graph6Encode(graph.InducedSubgraph(g, ci))
		graphs[g6] = struct{}{}
	}
	return len(graphs)
}

func checkOrbits(n int) int {
	less := func(i, j int) bool {
		if i < j && j < n {
			return true
		}
		if i%n == j%n {
			return i < j
		}
		return false
	}

	iter := itertools.TopologicalSorts(2*n, less)

	graphs := make(map[string]struct{})
	seen := make([]bool, n)
	g := graph.NewDense(n, nil)
	value := make([]int, 2*n)
	init := make([]int, 2*n)
	labels := make([]int, n)
	for iter.Next() {
		v := iter.Value()
		for i := range value {
			value[i] = v[i] % n
		}
		copy(init, value)
		if !isBest(value, init, labels) {
			continue
		}
		//Reset the graph g.
		for i := range g.Edges {
			g.Edges[i] = 0
		}
		for i := range g.DegreeSequence {
			g.DegreeSequence[i] = 0
		}
		g.NumberOfEdges = 0

		for i, u := range v {
			for j := range seen {
				seen[j] = false
			}
			for j := i + 1; j < 2*n; j++ {
				w := v[j] % n
				if w == u {
					for k := range seen {
						if seen[k] {
							g.AddEdge(u, k)
						}
					}
					break
				}
				seen[w] = !seen[w]
			}
		}
		ci := graph.CanonicalIsomorph(g)
		g6 := graph.Graph6Encode(graph.InducedSubgraph(g, ci))
		graphs[g6] = struct{}{}
	}
	return len(graphs)
}

//TODO Preallocate best.
func isBest(value, init, labels []int) bool {
	n := len(value) / 2
	numSeen := 0

	for i := 0; i < 2*n-1; i++ {
		//Reset the labels
		for i := range labels {
			labels[i] = -1
		}
		numSeen = 0
		//Rotate
		tmp := value[0]
		for i := 0; i < 2*n-1; i++ {
			x := value[i+1]
			if labels[x] == -1 {
				labels[x] = numSeen
				numSeen++
			}
			value[i] = labels[x]
		}
		value[2*n-1] = labels[tmp]
		if ints.Compare(init, value) == -1 {
			return false
		}
	}
	return true
}

func checkAllGraphs(n int) int {
	//Allocate the structures for the circle graph test and reuse them.
	mat := Zeros(0, 0)
	b := make([]byte, 0)
	neighbours := make([]uint, n)

	c := make(chan *graph.DenseGraph)
	go search.All(n, c, 0, 1)
	counter := 0
	for g := range c {
		if isCircleGraph(g, mat, b, neighbours) {
			counter++
		}
	}
	return counter
}

//enumerateParallel enumerates all circle graphs on n vertices using numWorkers goroutines and sends them to output.
func checkCanonicalDeletion(n int) int {
	output := make(chan *graph.DenseGraph)

	mat := Zeros(0, 0)
	b := make([]byte, 0)
	neighbours := make([]uint, n)

	go search.WithPruning(n, output, 0, 1, func(g *graph.DenseGraph) bool { return false }, func(g *graph.DenseGraph) bool { return !isCircleGraph(g, mat, b, neighbours) })

	counter := 0
	for range output {
		counter++
	}

	return counter
}

//isCircleGraph checks if the graph g is a circle graph. It will crash if n is larger than the number of bits in uint.
//It will use the space allocated in mat, b and neighbours. neighbours must be of length at least n.
//The method used is in Naji, Walid. "Reconnaissance des graphes de cordes." Discrete Mathematics 54.3 (1985): 329-337.
func isCircleGraph(g *graph.DenseGraph, mat *Matrix, b []byte, neighbours []uint) bool {

	components := graph.ConnectedComponents(g)

	for _, comp := range components {
		n := len(comp)

		numVars := n * n //We will have variables for (i,i) even if we don't need them.
		mat.N = numVars
		mat.M = 0
		for i := range mat.Entries {
			mat.Entries[i] = 0
		}
		mat.Entries = mat.Entries[:0]
		mat.RowPerm = mat.RowPerm[:0]
		numEqns := 0
		b = b[:0]
		for i := range neighbours {
			neighbours[i] = 0
		}
		neighbours = neighbours[:n]
		for i, v := range comp {
			for j := i + 1; j < n; j++ {
				u := comp[j]
				if g.IsEdge(v, u) {
					neighbours[i] |= (1 << uint(j))
					neighbours[j] |= (1 << uint(i))
				}
			}
		}

		for i := 0; i < n; i++ {
			v := comp[i]
			for j := i + 1; j < n; j++ {
				u := comp[j]
				if g.IsEdge(v, u) {
					//Find the vertices which are non-neighbours of both i and j.
					c := (^neighbours[i]) & (^neighbours[j])
					//Restrict to the first n bits.
					c &= (1 << uint(n)) - 1

					mat.AddRows(bits.OnesCount(c) + 1)
					//First type of equation
					mat.SetEntry(numEqns, i*n+j, 1)
					mat.SetEntry(numEqns, j*n+i, 1)
					b = append(b, 1)
					numEqns++

					//Other type of equation

					for c != 0 {
						y := c & -c
						v := bits.TrailingZeros(c)
						c ^= y

						mat.SetEntry(numEqns, v*n+i, 1)
						mat.SetEntry(numEqns, v*n+j, 1)
						b = append(b, 0)
						numEqns++
					}
					continue
				}
				c := neighbours[i] & neighbours[j]
				mat.AddRows(bits.OnesCount(c))
				for c != 0 {
					y := c & -c
					v := bits.TrailingZeros(c)
					c ^= y
					mat.SetEntry(numEqns, v*n+i, 1)
					mat.SetEntry(numEqns, v*n+j, 1)
					mat.SetEntry(numEqns, i*n+j, 1)
					mat.SetEntry(numEqns, j*n+i, 1)
					b = append(b, 1)
					numEqns++
				}
			}
		}
		if mat.HasSolution(b) == false {
			return false
		}
	}
	return true
}

//Matrix is a M x N dense matrix over F2.
//Each entry in the matrix is a bit and they are ordered left to right, top to bottom. The bits in a row are split into blocks of length 64 and each block is reversed before being encoded in a uint64.
//The entries are indexed from 0.
type Matrix struct {
	M       int
	N       int
	Entries []uint64
	RowPerm []int //The ith row in the current state of the matrix is the RowPerm[i]th row according to entries.
}

//Zeros initialises an M x N matrix which is all zeros.
func Zeros(M, N int) *Matrix {
	entries := make([]uint64, M*((N+63)/64))
	rowPerm := make([]int, M)
	for i := range rowPerm {
		rowPerm[i] = i
	}
	return &Matrix{M: M, N: N, Entries: entries, RowPerm: rowPerm}
}

//AddRows adds numRows extra zero rows to the bottom of the matrix.
func (m *Matrix) AddRows(numRows int) {
	width := (m.N + 63) / 64
	numNeeded := (m.M + numRows) * width
	if cap(m.Entries) >= numNeeded {
		m.Entries = m.Entries[:numNeeded]
		for i := m.M * width; i < numNeeded; i++ {
			m.Entries[i] = 0
		}
	} else {
		tmp := make([]uint64, numNeeded)
		copy(tmp, m.Entries)
		m.Entries = tmp
	}
	numNeeded = (m.M + numRows)
	if cap(m.RowPerm) >= numNeeded {
		m.RowPerm = m.RowPerm[:numNeeded]
		for i := m.M; i < numNeeded; i++ {
			m.RowPerm[i] = i
		}
	} else {
		tmp := make([]int, numNeeded)
		copy(tmp, m.RowPerm)
		m.RowPerm = tmp
		for i := m.M; i < numNeeded; i++ {
			m.RowPerm[i] = i
		}
	}
	m.M += numRows
}

//SetEntry sets the (i,j) entry to the value b.
//The value b must be either 0 or 1.
func (m *Matrix) SetEntry(i, j int, b byte) {
	r := m.RowPerm[i]
	switch b {
	case 0:
		m.Entries[r*((m.N+63)/64)+j/64] &^= (1 << (j % 64))
		return
	case 1:
		m.Entries[r*((m.N+63)/64)+j/64] |= (1 << (j % 64))
		return
	}
	panic("b is not 0 or 1")
}

//GetEntry returns the (i,j) entry of m.
func (m *Matrix) GetEntry(i, j int) uint64 {
	r := m.RowPerm[i]
	return (m.Entries[r*((m.N+63)/64)+j/64] >> (j % 64)) & 1
}

//AddRowTo replaces the dst row with dst row + src row.
func (m *Matrix) AddRowTo(src, dst int) {
	rSrc := m.RowPerm[src]
	rDst := m.RowPerm[dst]
	width := (m.N + 63) / 64
	for k := 0; k < width; k++ {
		m.Entries[rDst*width+k] ^= m.Entries[rSrc*width+k]
	}
}

//SwapRows swaps the rows i and j.
func (m *Matrix) SwapRows(i, j int) {
	m.RowPerm[i], m.RowPerm[j] = m.RowPerm[j], m.RowPerm[i]
}

//Copy returns a deep copy of the matrix.
func (m *Matrix) Copy() *Matrix {
	entries := make([]uint64, len(m.Entries))
	copy(entries, m.Entries)
	rowPerm := make([]int, m.M)
	copy(rowPerm, m.RowPerm)
	return &Matrix{M: m.M, N: m.N, Entries: entries, RowPerm: rowPerm}
}

//HasSolution checks if the equation m*x = b has a solution over F_2^N using dense Gaussian Elimination.
//This modifies the matrix m.
func (m *Matrix) HasSolution(b []byte) bool {
	r := 0
	c := 0
mainLoop:
	for r < m.M && c < m.N {
		//Find a pivot row
		for i := r; i < m.M; i++ {
			if m.GetEntry(i, c) == 1 {
				//Swap the rows
				m.SwapRows(i, r)
				b[i], b[r] = b[r], b[i]
				for k := r + 1; k < m.M; k++ {
					if m.GetEntry(k, c) == 1 {
						m.AddRowTo(r, k)
						b[k] ^= b[r]
					}
				}
				//Next Iteration
				r++
				c++
				continue mainLoop
			}
		}
		//No pivot found. Skip this column and repeat.
		c++
	}
	// fmt.Println(m, b)
	for i := r; i < m.M; i++ {
		if b[i] != 0 {
			return false
		}
	}
	return true
}

func (m Matrix) String() string {
	var s strings.Builder
	for i := 0; i < m.M; i++ {
		for j := 0; j < m.N; j++ {
			err := s.WriteByte(48 + byte(m.GetEntry(i, j)))
			if err != nil {
				panic(err)
			}
		}
		err := s.WriteByte(10)
		if err != nil {
			panic(err)
		}
	}
	return s.String()
}
